@@ -1,7 +1,7 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009, 2010  Université de Bordeaux 1
- * Copyright (C) 2010  Centre National de la Recherche Scientifique
+ * Copyright (C) 2009, 2010, 2014-2015  Université de Bordeaux
+ * Copyright (C) 2010  CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -15,6 +15,7 @@
  * See the GNU Lesser General Public License in COPYING.LGPL for more details.
  */
 
+/* Complex float macros */
 
 #include <complex.h>
 #include "blas_complex.h"
@@ -24,15 +25,8 @@
 #define STARPU_LU(name)       starpu_clu_##name
 #define COMPLEX_LU
 
-#ifdef STARPU_HAVE_MAGMA
-#include <magmablas.h>
-#define CUBLAS_GEMM	magmablas_cgemm
-#define CUBLAS_TRSM	magmablas_ctrsm
-#else
 #define CUBLAS_GEMM	cublasCgemm
 #define CUBLAS_TRSM	cublasCtrsm
-#endif
-
 #define CUBLAS_SCAL	cublasCscal
 #define CUBLAS_GER	cublasCgeru
 #define CUBLAS_SWAP	cublasCswap
