@@ -1,7 +1,7 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009-2013  Université de Bordeaux 1
- * Copyright (C) 2010, 2011  Centre National de la Recherche Scientifique
+ * Copyright (C) 2009-2013, 2016  Université de Bordeaux
+ * Copyright (C) 2010, 2011  CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -36,7 +36,7 @@ void _starpu_open_debug_logfile(void);
 void _starpu_close_debug_logfile(void);
 
 /* Write into StarPU's log file */
-void _starpu_print_to_logfile(const char *format, ...);
+void _starpu_print_to_logfile(const char *format, ...) STARPU_ATTRIBUTE_FORMAT(printf, 1, 2);
 
 /* Tell gdb whether FXT is compiled in or not */
 extern int _starpu_use_fxt;
