@@ -1,9 +1,9 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009-2013  Université de Bordeaux 1
- * Copyright (C) 2010, 2011, 2012, 2013  Centre National de la Recherche Scientifique
+ * Copyright (C) 2009-2015  Université de Bordeaux
+ * Copyright (C) 2010, 2011, 2012, 2013, 2014, 2015, 2016  CNRS
  * Copyright (C) 2011  Télécom-SudParis
- * Copyright (C) 2011  INRIA
+ * Copyright (C) 2011, 2014  INRIA
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -24,6 +24,7 @@
 #include <core/jobs.h>
 #include <core/task.h>
 #include <core/task_bundle.h>
+#include <core/dependencies/data_concurrency.h>
 #include <common/config.h>
 #include <common/utils.h>
 #include <common/fxt.h>
@@ -32,22 +33,40 @@
 #include <math.h>
 #include <string.h>
 #include <core/debug.h>
+#include <core/sched_ctx.h>
 #include <time.h>
+#include <signal.h>
 #ifdef STARPU_HAVE_WINDOWS
 #include <windows.h>
 #endif
-#include <core/sched_ctx.h>
 
 /* XXX this should be reinitialized when StarPU is shutdown (or we should make
  * sure that no task remains !) */
 /* TODO we could make this hierarchical to avoid contention ? */
-static starpu_pthread_cond_t submitted_cond = STARPU_PTHREAD_COND_INITIALIZER;
+//static starpu_pthread_cond_t submitted_cond = STARPU_PTHREAD_COND_INITIALIZER;
 
 /* This key stores the task currently handled by the thread, note that we
  * cannot use the worker structure to store that information because it is
  * possible that we have a task with a NULL codelet, which means its callback
  * could be executed by a user thread as well. */
 static starpu_pthread_key_t current_task_key;
+static int limit_min_submitted_tasks;
+static int limit_max_submitted_tasks;
+static int watchdog_crash;
+
+/* Called once at starpu_init */
+void _starpu_task_init(void)
+{
+	STARPU_PTHREAD_KEY_CREATE(&current_task_key, NULL);
+	limit_min_submitted_tasks = starpu_get_env_number("STARPU_LIMIT_MIN_SUBMITTED_TASKS");
+	limit_max_submitted_tasks = starpu_get_env_number("STARPU_LIMIT_MAX_SUBMITTED_TASKS");
+	watchdog_crash = starpu_get_env_number("STARPU_WATCHDOG_CRASH");
+}
+
+void _starpu_task_deinit(void)
+{
+	STARPU_PTHREAD_KEY_DELETE(current_task_key);
+}
 
 void starpu_task_init(struct starpu_task *task)
 {
@@ -81,9 +100,11 @@ void starpu_task_init(struct starpu_task *task)
 	task->flops = 0.0;
 
 	task->scheduled = 0;
+	task->prefetched = 0;
 
 	task->dyn_handles = NULL;
 	task->dyn_interfaces = NULL;
+	task->name = NULL;
 }
 
 /* Free all the ressources allocated for a task, without deallocating the task
@@ -114,6 +135,12 @@ void starpu_task_clean(struct starpu_task *task)
 		task->dyn_interfaces = NULL;
 	}
 
+	if (task->dyn_modes)
+	{
+		free(task->dyn_modes);
+		task->dyn_modes = NULL;
+	}
+
 	struct _starpu_job *j = (struct _starpu_job *)task->starpu_private;
 
 	if (j)
@@ -170,6 +197,10 @@ void _starpu_task_destroy(struct starpu_task *task)
 		if (task->prologue_callback_arg_free)
 			free(task->prologue_callback_arg);
 
+		/* Does user want StarPU release prologue_pop_arg ? */
+		if (task->prologue_callback_pop_arg_free)
+			free(task->prologue_callback_pop_arg);
+
 		free(task);
 	}
 }
@@ -181,6 +212,13 @@ void starpu_task_destroy(struct starpu_task *task)
 	_starpu_task_destroy(task);
 }
 
+int starpu_task_finished(struct starpu_task *task)
+{
+	STARPU_ASSERT(task);
+	STARPU_ASSERT_MSG(!task->detach, "starpu_task_finished can only be called on tasks with detach = 0");
+	return _starpu_job_finished(_starpu_get_job_associated_to_task(task));
+}
+
 int starpu_task_wait(struct starpu_task *task)
 {
         _STARPU_LOG_IN();
@@ -190,19 +228,17 @@ int starpu_task_wait(struct starpu_task *task)
 
 	if (task->detach || task->synchronous)
 	{
-		_STARPU_DEBUG("Task is detached or asynchronous. Waiting returns immediately\n");
+		_STARPU_DEBUG("Task is detached or synchronous. Waiting returns immediately\n");
 		_STARPU_LOG_OUT_TAG("einval");
 		return -EINVAL;
 	}
 
-	if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-	{
-		_STARPU_LOG_OUT_TAG("edeadlk");
-		return -EDEADLK;
-	}
+	STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "starpu_task_wait must not be called from a task or callback");
 
 	struct _starpu_job *j = (struct _starpu_job *)task->starpu_private;
 
+	_STARPU_TRACE_TASK_WAIT_START(j);
+
 	_starpu_wait_job(j);
 
 	/* as this is a synchronous task, the liberation of the job
@@ -210,49 +246,80 @@ int starpu_task_wait(struct starpu_task *task)
 	if (task->destroy)
 		_starpu_task_destroy(task);
 
+	_STARPU_TRACE_TASK_WAIT_END();
         _STARPU_LOG_OUT();
 	return 0;
 }
 
-struct _starpu_job *_starpu_get_job_associated_to_task(struct starpu_task *task)
+#ifdef STARPU_OPENMP
+int _starpu_task_test_termination(struct starpu_task *task)
 {
 	STARPU_ASSERT(task);
+	STARPU_ASSERT_MSG(!task->detach, "starpu_task_wait can only be called on tasks with detach = 0");
 
-	if (!task->starpu_private)
+	if (task->detach || task->synchronous)
 	{
-		struct _starpu_job *j = _starpu_job_create(task);
-		task->starpu_private = j;
+		_STARPU_DEBUG("Task is detached or synchronous\n");
+		_STARPU_LOG_OUT_TAG("einval");
+		return -EINVAL;
 	}
 
-	return (struct _starpu_job *)task->starpu_private;
+	struct _starpu_job *j = (struct _starpu_job *)task->starpu_private;
+
+	int ret = _starpu_test_job_termination(j);
+
+	if (ret)
+	{
+		if (task->destroy)
+			_starpu_task_destroy(task);
+	}
+
+	return ret;
 }
+#endif
 
 /* NB in case we have a regenerable task, it is possible that the job was
  * already counted. */
 int _starpu_submit_job(struct _starpu_job *j)
 {
-
 	struct starpu_task *task = j->task;
+	int ret;
+#ifdef STARPU_OPENMP
+	const unsigned continuation = j->continuation;
+#else
+	const unsigned continuation = 0;
+#endif
 
 	_STARPU_LOG_IN();
 	/* notify bound computation of a new task */
 	_starpu_bound_record(j);
 
 	_starpu_increment_nsubmitted_tasks_of_sched_ctx(j->task->sched_ctx);
+	_starpu_sched_task_submit(task);
 
 #ifdef STARPU_USE_SC_HYPERVISOR
 	struct _starpu_sched_ctx *sched_ctx = _starpu_get_sched_ctx_struct(j->task->sched_ctx);
 	if(sched_ctx != NULL && j->task->sched_ctx != _starpu_get_initial_sched_ctx()->id && j->task->sched_ctx != STARPU_NMAX_SCHED_CTXS
 	   && sched_ctx->perf_counters != NULL)
 	{
-		_starpu_compute_buffers_footprint(j->task->cl->model, STARPU_CPU_DEFAULT, 0, j);
-		int i;
+		struct starpu_perfmodel_arch arch;
+		arch.devices = (struct starpu_perfmodel_device*)malloc(sizeof(struct starpu_perfmodel_device));
+		arch.ndevices = 1;
+		arch.devices[0].type = STARPU_CPU_WORKER;
+		arch.devices[0].devid = 0;
+		arch.devices[0].ncores = 1;
+		_starpu_compute_buffers_footprint(j->task->cl->model, &arch, 0, j);
+		free(arch.devices);
 		size_t data_size = 0;
-		for(i = 0; i < STARPU_NMAXBUFS; i++)
+		if (j->task->cl)
 		{
-			starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, i);
-			if (handle != NULL)
-				data_size += _starpu_data_get_size(handle);
+			unsigned i, nbuffers = STARPU_TASK_GET_NBUFFERS(j->task);
+			for(i = 0; i < nbuffers; i++)
+			{
+				starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, i);
+				if (handle != NULL)
+					data_size += _starpu_data_get_size(handle);
+			}
 		}
 
 		_STARPU_TRACE_HYPERVISOR_BEGIN();
@@ -262,10 +329,11 @@ int _starpu_submit_job(struct _starpu_job *j)
 #endif//STARPU_USE_SC_HYPERVISOR
 
 	/* We retain handle reference count */
-	if (task->cl)
+	if (task->cl && !continuation)
 	{
 		unsigned i;
-		for (i=0; i<task->cl->nbuffers; i++)
+		unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+		for (i=0; i<nbuffers; i++)
 		{
 			starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, i);
 			_starpu_spin_lock(&handle->header_lock);
@@ -276,15 +344,26 @@ int _starpu_submit_job(struct _starpu_job *j)
 
 	STARPU_PTHREAD_MUTEX_LOCK(&j->sync_mutex);
 
-	/* Need to atomically set submitted to 1 and check dependencies, since
-	 * this is concucrent with _starpu_notify_cg */
-	j->terminated = 0;
-	if (!j->submitted)
-		j->submitted = 1;
-	else
-		j->submitted = 2;
+	_starpu_handle_job_submission(j);
 
-	int ret = _starpu_enforce_deps_and_schedule(j);
+#ifdef STARPU_OPENMP
+	if (continuation)
+	{
+		j->discontinuous = 1;
+		j->continuation  = 0;
+	}
+#endif
+
+#ifdef STARPU_OPENMP
+	if (continuation)
+	{
+		ret = _starpu_reenforce_task_deps_and_schedule(j);
+	}
+	else
+#endif
+	{
+		ret = _starpu_enforce_deps_and_schedule(j);
+	}
 
 	_STARPU_LOG_OUT();
 	return ret;
@@ -298,6 +377,7 @@ void _starpu_codelet_check_deprecated_fields(struct starpu_codelet *cl)
 		return;
 
 	int is_where_unset = cl->where == 0;
+	unsigned i, some_impl;
 
 	/* Check deprecated and unset fields (where, <device>_func,
  	 * <device>_funcs) */
@@ -313,11 +393,18 @@ void _starpu_codelet_check_deprecated_fields(struct starpu_codelet *cl)
 		cl->cpu_funcs[0] = cl->cpu_func;
 		cl->cpu_func = STARPU_MULTIPLE_CPU_IMPLEMENTATIONS;
 	}
-	if (cl->cpu_funcs[0] && cl->cpu_func == 0)
+	some_impl = 0;
+	for (i = 0; i < STARPU_MAXIMPLEMENTATIONS; i++)
+		if (cl->cpu_funcs[i])
+		{
+			some_impl = 1;
+			break;
+		}
+	if (some_impl && cl->cpu_func == 0)
 	{
 		cl->cpu_func = STARPU_MULTIPLE_CPU_IMPLEMENTATIONS;
 	}
-	if (cl->cpu_funcs[0] && is_where_unset)
+	if (some_impl && is_where_unset)
 	{
 		cl->where |= STARPU_CPU;
 	}
@@ -333,11 +420,18 @@ void _starpu_codelet_check_deprecated_fields(struct starpu_codelet *cl)
 		cl->cuda_funcs[0] = cl->cuda_func;
 		cl->cuda_func = STARPU_MULTIPLE_CUDA_IMPLEMENTATIONS;
 	}
-	if (cl->cuda_funcs[0] && cl->cuda_func == 0)
+	some_impl = 0;
+	for (i = 0; i < STARPU_MAXIMPLEMENTATIONS; i++)
+		if (cl->cuda_funcs[i])
+		{
+			some_impl = 1;
+			break;
+		}
+	if (some_impl && cl->cuda_func == 0)
 	{
 		cl->cuda_func = STARPU_MULTIPLE_CUDA_IMPLEMENTATIONS;
 	}
-	if (cl->cuda_funcs[0] && is_where_unset)
+	if (some_impl && is_where_unset)
 	{
 		cl->where |= STARPU_CUDA;
 	}
@@ -353,52 +447,67 @@ void _starpu_codelet_check_deprecated_fields(struct starpu_codelet *cl)
 		cl->opencl_funcs[0] = cl->opencl_func;
 		cl->opencl_func = STARPU_MULTIPLE_OPENCL_IMPLEMENTATIONS;
 	}
-	if (cl->opencl_funcs[0] && cl->opencl_func == 0)
+	some_impl = 0;
+	for (i = 0; i < STARPU_MAXIMPLEMENTATIONS; i++)
+		if (cl->opencl_funcs[i])
+		{
+			some_impl = 1;
+			break;
+		}
+	if (some_impl && cl->opencl_func == 0)
 	{
 		cl->opencl_func = STARPU_MULTIPLE_OPENCL_IMPLEMENTATIONS;
 	}
-	if (cl->opencl_funcs[0] && is_where_unset)
+	if (some_impl && is_where_unset)
 	{
 		cl->where |= STARPU_OPENCL;
 	}
-}
 
-void _starpu_task_check_deprecated_fields(struct starpu_task *task)
-{
-	if (task->cl)
-	{
-		unsigned i;
-		for(i=0; i<STARPU_MIN(task->cl->nbuffers, STARPU_NMAXBUFS) ; i++)
+	some_impl = 0;
+	for (i = 0; i < STARPU_MAXIMPLEMENTATIONS; i++)
+		if (cl->mic_funcs[i])
 		{
-			if (task->buffers[i].handle && task->handles[i])
-			{
-				_STARPU_DISP("[warning][struct starpu_task] task->buffers[%u] and task->handles[%u] both set. Ignoring task->buffers[%u] ?\n", i, i, i);
-				STARPU_ASSERT(task->buffers[i].mode == task->cl->modes[i]);
-				STARPU_ABORT();
-			}
-			if (task->buffers[i].handle)
-			{
-				task->handles[i] = task->buffers[i].handle;
-				task->cl->modes[i] = task->buffers[i].mode;
-			}
+			some_impl = 1;
+			break;
 		}
+	if (some_impl && is_where_unset)
+	{
+		cl->where |= STARPU_MIC;
+	}
+
+	some_impl = 0;
+	for (i = 0; i < STARPU_MAXIMPLEMENTATIONS; i++)
+		if (cl->scc_funcs[i])
+		{
+			some_impl = 1;
+			break;
+		}
+	if (some_impl && is_where_unset)
+	{
+		cl->where |= STARPU_SCC;
+	}
+
+	some_impl = 0;
+	for (i = 0; i < STARPU_MAXIMPLEMENTATIONS; i++)
+		if (cl->cpu_funcs_name[i])
+		{
+			some_impl = 1;
+			break;
+		}
+	if (some_impl && is_where_unset)
+	{
+		cl->where |= STARPU_MIC|STARPU_SCC;
 	}
 }
 
-/* application should submit new tasks to StarPU through this function */
-int starpu_task_submit(struct starpu_task *task)
+void _starpu_task_check_deprecated_fields(struct starpu_task *task STARPU_ATTRIBUTE_UNUSED)
 {
-	_STARPU_LOG_IN();
-	STARPU_ASSERT(task);
-	STARPU_ASSERT_MSG(task->magic == 42, "Tasks must be created with starpu_task_create, or initialized with starpu_task_init.");
+	/* None any more */
+}
 
-	int ret;
+static int _starpu_task_submit_head(struct starpu_task *task)
+{
 	unsigned is_sync = task->synchronous;
-	starpu_task_bundle_t bundle = task->bundle;
-
-	/* internally, StarPU manipulates a struct _starpu_job * which is a wrapper around a
-	* task structure, it is possible that this job structure was already
-	* allocated. */
 	struct _starpu_job *j = _starpu_get_job_associated_to_task(task);
 
 	if (j->internal)
@@ -416,12 +525,7 @@ int starpu_task_submit(struct starpu_task *task)
 	{
 		/* Perhaps it is not possible to submit a synchronous
 		 * (blocking) task */
-		if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-		{
-			_STARPU_LOG_OUT_TAG("EDEADLK");
-			return -EDEADLK;
-		}
-
+		STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "submitting a synchronous task must not be done from a task or a callback");
 		task->detach = 0;
 	}
 
@@ -431,17 +535,19 @@ int starpu_task_submit(struct starpu_task *task)
 	if (task->cl)
 	{
 		unsigned i;
+		unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+		_STARPU_TRACE_UPDATE_TASK_CNT(0);
 
 		/* Check buffers */
 		if (task->dyn_handles == NULL)
-			STARPU_ASSERT_MSG(task->cl->nbuffers <= STARPU_NMAXBUFS, "Codelet %p has too many buffers (%d vs max %d). Either use --enable-maxbuffers configure option to increase the max, or use dyn_handles instead of handles.", task->cl, task->cl->nbuffers, STARPU_NMAXBUFS);
+			STARPU_ASSERT_MSG(STARPU_TASK_GET_NBUFFERS(task) <= STARPU_NMAXBUFS, "Codelet %p has too many buffers (%d vs max %d). Either use --enable-maxbuffers configure option to increase the max, or use dyn_handles instead of handles.", task->cl, STARPU_TASK_GET_NBUFFERS(task), STARPU_NMAXBUFS);
 
 		if (task->dyn_handles)
 		{
-			task->dyn_interfaces = malloc(task->cl->nbuffers * sizeof(void *));
+			task->dyn_interfaces = malloc(nbuffers * sizeof(void *));
 		}
 
-		for (i = 0; i < task->cl->nbuffers; i++)
+		for (i = 0; i < nbuffers; i++)
 		{
 			starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, i);
 			/* Make sure handles are not partitioned */
@@ -468,16 +574,66 @@ int starpu_task_submit(struct starpu_task *task)
 			return -ENODEV;
 		}
 
-		_starpu_detect_implicit_data_deps(task);
+		if (task->cl->model)
+			_starpu_init_and_load_perfmodel(task->cl->model);
 
-
-		if (task->cl->model && task->cl->model->symbol)
-			_starpu_load_perfmodel(task->cl->model);
-
-		if (task->cl->power_model && task->cl->power_model->symbol)
-			_starpu_load_perfmodel(task->cl->power_model);
+		if (task->cl->energy_model)
+			_starpu_init_and_load_perfmodel(task->cl->energy_model);
 	}
 
+	return 0;
+}
+
+/* application should submit new tasks to StarPU through this function */
+int starpu_task_submit(struct starpu_task *task)
+{
+	_STARPU_LOG_IN();
+	STARPU_ASSERT(task);
+	STARPU_ASSERT_MSG(task->magic == 42, "Tasks must be created with starpu_task_create, or initialized with starpu_task_init.");
+
+	int ret;
+	unsigned is_sync = task->synchronous;
+	starpu_task_bundle_t bundle = task->bundle;
+	/* internally, StarPU manipulates a struct _starpu_job * which is a wrapper around a
+	* task structure, it is possible that this job structure was already
+	* allocated. */
+	struct _starpu_job *j = _starpu_get_job_associated_to_task(task);
+	const unsigned continuation =
+#ifdef STARPU_OPENMP
+		j->continuation
+#else
+		0
+#endif
+		;
+
+	_STARPU_TRACE_TASK_SUBMIT_START();
+
+	if (!j->internal)
+	{
+		int nsubmitted_tasks = starpu_task_nsubmitted();
+		if (limit_max_submitted_tasks >= 0 && limit_max_submitted_tasks < nsubmitted_tasks
+			&& limit_min_submitted_tasks >= 0 && limit_min_submitted_tasks < nsubmitted_tasks)
+		{
+			starpu_do_schedule();
+			starpu_task_wait_for_n_submitted(limit_min_submitted_tasks);
+		}
+	}
+
+
+	ret = _starpu_task_submit_head(task);
+	if (ret)
+	{
+		_STARPU_TRACE_TASK_SUBMIT_END();
+		return ret;
+	}
+
+	if (!j->internal && !continuation)
+		_STARPU_TRACE_TASK_SUBMIT(j);
+
+	/* If this is a continuation, we don't modify the implicit data dependencies detected earlier. */
+	if (task->cl && !continuation)
+		_starpu_detect_implicit_data_deps(task);
+
 	if (bundle)
 	{
 		/* We need to make sure that models for other tasks of the
@@ -490,11 +646,11 @@ int starpu_task_submit(struct starpu_task *task)
 
 		while (entry)
 		{
-			if (entry->task->cl->model && entry->task->cl->model->symbol)
-				_starpu_load_perfmodel(entry->task->cl->model);
+			if (entry->task->cl->model)
+				_starpu_init_and_load_perfmodel(entry->task->cl->model);
 
-			if (entry->task->cl->power_model && entry->task->cl->power_model->symbol)
-				_starpu_load_perfmodel(entry->task->cl->power_model);
+			if (entry->task->cl->energy_model)
+				_starpu_init_and_load_perfmodel(entry->task->cl->energy_model);
 
 			entry = entry->next;
 		}
@@ -513,19 +669,23 @@ int starpu_task_submit(struct starpu_task *task)
 	 * dependency. */
 	task->status = STARPU_TASK_BLOCKED;
 
-
 	if (profiling)
 		_starpu_clock_gettime(&info->submit_time);
 
 	ret = _starpu_submit_job(j);
+#ifdef STARPU_SIMGRID
+	MSG_process_sleep(0.000001);
+#endif
 
 	if (is_sync)
 	{
+		_starpu_sched_do_schedule(task->sched_ctx);
 		_starpu_wait_job(j);
 		if (task->destroy)
 		     _starpu_task_destroy(task);
 	}
 
+	_STARPU_TRACE_TASK_SUBMIT_END();
         _STARPU_LOG_OUT();
 	return ret;
 }
@@ -548,49 +708,22 @@ int starpu_task_submit_to_ctx(struct starpu_task *task, unsigned sched_ctx_id)
  * skipping dependencies completely (when it knows what it is doing).  */
 int _starpu_task_submit_nodeps(struct starpu_task *task)
 {
-	_starpu_task_check_deprecated_fields(task);
-	_starpu_codelet_check_deprecated_fields(task->cl);
-
-	if (task->cl)
-	{
-		if (task->cl->model)
-			_starpu_load_perfmodel(task->cl->model);
-
-		if (task->cl->power_model)
-			_starpu_load_perfmodel(task->cl->power_model);
-	}
+	int ret;
+	ret = _starpu_task_submit_head(task);
+	STARPU_ASSERT(ret == 0);
 
 	struct _starpu_job *j = _starpu_get_job_associated_to_task(task);
 
-	if (j->internal)
-	{
-		// Internal tasks are submitted to initial context
-		j->task->sched_ctx = _starpu_get_initial_sched_ctx()->id;
-	}
-	else if (task->sched_ctx == STARPU_NMAX_SCHED_CTXS)
-	{
-		// If the task has not specified a context, we set the current context
-		j->task->sched_ctx = _starpu_sched_ctx_get_current_context();
-	}
-
 	_starpu_increment_nsubmitted_tasks_of_sched_ctx(j->task->sched_ctx);
+	_starpu_sched_task_submit(task);
+
 	STARPU_PTHREAD_MUTEX_LOCK(&j->sync_mutex);
-
-	j->submitted = 1;
-
+	_starpu_handle_job_submission(j);
+	_starpu_increment_nready_tasks_of_sched_ctx(j->task->sched_ctx, j->task->flops, j->task);
 	if (task->cl)
-	{
 		/* This would be done by data dependencies checking */
-		unsigned i;
-		for (i=0 ; i<task->cl->nbuffers ; i++)
-		{
-			starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(j->task, i);
-			_STARPU_JOB_SET_ORDERED_BUFFER_HANDLE(j, handle, i);
-			enum starpu_data_access_mode mode = STARPU_CODELET_GET_MODE(j->task->cl, i);
-			_STARPU_JOB_SET_ORDERED_BUFFER_MODE(j, mode, i);
-		}
-	}
-
+		_starpu_job_set_ordered_buffers(j);
+	task->status = STARPU_TASK_READY;
 	STARPU_PTHREAD_MUTEX_UNLOCK(&j->sync_mutex);
 
 	return _starpu_push_task(j);
@@ -602,22 +735,17 @@ int _starpu_task_submit_nodeps(struct starpu_task *task)
 int _starpu_task_submit_conversion_task(struct starpu_task *task,
 					unsigned int workerid)
 {
+	int ret;
 	STARPU_ASSERT(task->cl);
 	STARPU_ASSERT(task->execute_on_a_specific_worker);
 
-	_starpu_task_check_deprecated_fields(task);
-	_starpu_codelet_check_deprecated_fields(task->cl);
+	ret = _starpu_task_submit_head(task);
+	STARPU_ASSERT(ret == 0);
 
-	/* We should factorize that */
-	if (task->cl->model)
-		_starpu_load_perfmodel(task->cl->model);
-
-	if (task->cl->power_model)
-		_starpu_load_perfmodel(task->cl->power_model);
-
-	/* We retain handle reference count */
+	/* We retain handle reference count that would have been acquired by data dependencies.  */
 	unsigned i;
-	for (i=0; i<task->cl->nbuffers; i++)
+	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+	for (i=0; i<nbuffers; i++)
 	{
 		starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, i);
 		_starpu_spin_lock(&handle->header_lock);
@@ -627,30 +755,13 @@ int _starpu_task_submit_conversion_task(struct starpu_task *task,
 
 	struct _starpu_job *j = _starpu_get_job_associated_to_task(task);
 
-	if (j->internal)
-	{
-		// Internal tasks are submitted to initial context
-		j->task->sched_ctx = _starpu_get_initial_sched_ctx()->id;
-	}
-	else if (task->sched_ctx == STARPU_NMAX_SCHED_CTXS)
-	{
-		// If the task has not specified a context, we set the current context
-		j->task->sched_ctx = _starpu_sched_ctx_get_current_context();
-	}
-
 	_starpu_increment_nsubmitted_tasks_of_sched_ctx(j->task->sched_ctx);
-	STARPU_PTHREAD_MUTEX_LOCK(&j->sync_mutex);
-	j->submitted = 1;
-	_starpu_increment_nready_tasks_of_sched_ctx(j->task->sched_ctx, j->task->flops);
-	for (i=0 ; i<task->cl->nbuffers ; i++)
-	{
-		starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(j->task, i);
-		_STARPU_JOB_SET_ORDERED_BUFFER_HANDLE(j, handle, i);
-		enum starpu_data_access_mode mode = STARPU_CODELET_GET_MODE(j->task->cl, i);
-		_STARPU_JOB_SET_ORDERED_BUFFER_MODE(j, mode, i);
-	}
+	_starpu_sched_task_submit(task);
 
-        _STARPU_LOG_IN();
+	STARPU_PTHREAD_MUTEX_LOCK(&j->sync_mutex);
+	_starpu_handle_job_submission(j);
+	_starpu_increment_nready_tasks_of_sched_ctx(j->task->sched_ctx, j->task->flops, j->task);
+	_starpu_job_set_ordered_buffers(j);
 
 	task->status = STARPU_TASK_READY;
 	_starpu_profiling_set_task_push_start_time(task);
@@ -665,9 +776,7 @@ int _starpu_task_submit_conversion_task(struct starpu_task *task,
 
 	_starpu_profiling_set_task_push_end_time(task);
 
-        _STARPU_LOG_OUT();
 	STARPU_PTHREAD_MUTEX_UNLOCK(&j->sync_mutex);
-
 	return 0;
 }
 
@@ -705,7 +814,7 @@ void starpu_codelet_display_stats(struct starpu_codelet *cl)
  * regenerable is not considered finished until it was explicitely set as
  * non-regenerale anymore (eg. from a callback).
  */
-int starpu_task_wait_for_all(void)
+int _starpu_task_wait_for_all_and_return_nb_waited_tasks(void)
 {
 	unsigned nsched_ctxs = _starpu_get_nsched_ctxs();
 	unsigned sched_ctx_id = nsched_ctxs == 1 ? 0 : starpu_sched_ctx_get_context();
@@ -715,25 +824,99 @@ int starpu_task_wait_for_all(void)
 	if (sched_ctx_id == STARPU_NMAX_SCHED_CTXS)
 	{
 		_STARPU_DEBUG("Waiting for all tasks\n");
-		if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-			return -EDEADLK;
+		STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "starpu_task_wait_for_all must not be called from a task or callback");
 
 #ifdef HAVE_AYUDAME_H
 		if (AYU_event) AYU_event(AYU_BARRIER, 0, NULL);
 #endif
 		struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
 		if(config->topology.nsched_ctxs == 1)
-			starpu_task_wait_for_all_in_ctx(0);
+		{
+			_starpu_sched_do_schedule(0);
+			return _starpu_task_wait_for_all_in_ctx_and_return_nb_waited_tasks(0);
+		}
 		else
 		{
 			int s;
 			for(s = 0; s < STARPU_NMAX_SCHED_CTXS; s++)
+			{
+				if(config->sched_ctxs[s].id != STARPU_NMAX_SCHED_CTXS)
+				{
+					_starpu_sched_do_schedule(config->sched_ctxs[s].id);
+				}
+			}
+			for(s = 0; s < STARPU_NMAX_SCHED_CTXS; s++)
 			{
 				if(config->sched_ctxs[s].id != STARPU_NMAX_SCHED_CTXS)
 				{
 					starpu_task_wait_for_all_in_ctx(config->sched_ctxs[s].id);
 				}
 			}
+			return 0;
+		}
+	}
+	else
+	{
+		_starpu_sched_do_schedule(sched_ctx_id);
+		_STARPU_DEBUG("Waiting for tasks submitted to context %u\n", sched_ctx_id);
+		return _starpu_task_wait_for_all_in_ctx_and_return_nb_waited_tasks(sched_ctx_id);
+	}
+}
+
+int starpu_task_wait_for_all(void)
+{
+	_starpu_task_wait_for_all_and_return_nb_waited_tasks();
+	return 0;
+}
+
+int _starpu_task_wait_for_all_in_ctx_and_return_nb_waited_tasks(unsigned sched_ctx)
+{
+	_STARPU_TRACE_TASK_WAIT_FOR_ALL_START();
+	int ret = _starpu_wait_for_all_tasks_of_sched_ctx(sched_ctx);
+	_STARPU_TRACE_TASK_WAIT_FOR_ALL_END();
+#ifdef HAVE_AYUDAME_H
+	/* TODO: improve Temanejo into knowing about contexts ... */
+	if (AYU_event) AYU_event(AYU_BARRIER, 0, NULL);
+#endif
+	return ret;
+}
+
+int starpu_task_wait_for_all_in_ctx(unsigned sched_ctx)
+{
+	_starpu_task_wait_for_all_in_ctx_and_return_nb_waited_tasks(sched_ctx);
+	return 0;
+}
+
+/*
+ * We wait until there's a certain number of the tasks that have already been
+ * submitted left. Note that a regenerable is not considered finished until it
+ * was explicitely set as non-regenerale anymore (eg. from a callback).
+ */
+int starpu_task_wait_for_n_submitted(unsigned n)
+{
+	unsigned nsched_ctxs = _starpu_get_nsched_ctxs();
+	unsigned sched_ctx_id = nsched_ctxs == 1 ? 0 : starpu_sched_ctx_get_context();
+
+	/* if there is no indication about which context to wait,
+	   we wait for all tasks submitted to starpu */
+	if (sched_ctx_id == STARPU_NMAX_SCHED_CTXS)
+	{
+		_STARPU_DEBUG("Waiting for all tasks\n");
+		STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "starpu_task_wait_for_n_submitted must not be called from a task or callback");
+
+		struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
+		if(config->topology.nsched_ctxs == 1)
+			_starpu_wait_for_n_submitted_tasks_of_sched_ctx(0, n);
+		else
+		{
+			int s;
+			for(s = 0; s < STARPU_NMAX_SCHED_CTXS; s++)
+			{
+				if(config->sched_ctxs[s].id != STARPU_NMAX_SCHED_CTXS)
+				{
+					_starpu_wait_for_n_submitted_tasks_of_sched_ctx(config->sched_ctxs[s].id, n);
+				}
+			}
 		}
 
 		return 0;
@@ -741,21 +924,14 @@ int starpu_task_wait_for_all(void)
 	else
 	{
 		_STARPU_DEBUG("Waiting for tasks submitted to context %u\n", sched_ctx_id);
-		_starpu_wait_for_all_tasks_of_sched_ctx(sched_ctx_id);
-#ifdef HAVE_AYUDAME_H
-		/* TODO: improve Temanejo into knowing about contexts ... */
-		if (AYU_event) AYU_event(AYU_BARRIER, 0, NULL);
-#endif
+		_starpu_wait_for_n_submitted_tasks_of_sched_ctx(sched_ctx_id, n);
 	}
 	return 0;
 }
 
-int starpu_task_wait_for_all_in_ctx(unsigned sched_ctx)
+int starpu_task_wait_for_n_submitted_in_ctx(unsigned sched_ctx, unsigned n)
 {
-	_starpu_wait_for_all_tasks_of_sched_ctx(sched_ctx);
-#ifdef HAVE_AYUDAME_H
-	if (AYU_event) AYU_event(AYU_BARRIER, 0, NULL);
-#endif
+	_starpu_wait_for_n_submitted_tasks_of_sched_ctx(sched_ctx, n);
 
 	return 0;
 }
@@ -765,8 +941,7 @@ int starpu_task_wait_for_all_in_ctx(unsigned sched_ctx)
  */
 int starpu_task_wait_for_no_ready(void)
 {
-	if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-		return -EDEADLK;
+	STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "starpu_task_wait_for_no_ready must not be called from a task or callback");
 
 	struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
 	if(config->topology.nsched_ctxs == 1)
@@ -786,6 +961,24 @@ int starpu_task_wait_for_no_ready(void)
 	return 0;
 }
 
+void starpu_do_schedule(void)
+{
+	struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
+	if(config->topology.nsched_ctxs == 1)
+		_starpu_sched_do_schedule(0);
+	else
+	{
+		int s;
+		for(s = 0; s < STARPU_NMAX_SCHED_CTXS; s++)
+		{
+			if(config->sched_ctxs[s].id != STARPU_NMAX_SCHED_CTXS)
+			{
+				_starpu_sched_do_schedule(config->sched_ctxs[s].id);
+			}
+		}
+	}
+}
+
 void
 starpu_drivers_request_termination(void)
 {
@@ -839,7 +1032,7 @@ int starpu_task_nready(void)
 	int nready = 0;
 	struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
 	if(config->topology.nsched_ctxs == 1)
-		nready = starpu_get_nready_tasks_of_sched_ctx(0);
+		nready = starpu_sched_ctx_get_nready_tasks(0);
 	else
 	{
 		int s;
@@ -847,7 +1040,7 @@ int starpu_task_nready(void)
 		{
 			if(config->sched_ctxs[s].id != STARPU_NMAX_SCHED_CTXS)
 			{
-				nready += starpu_get_nready_tasks_of_sched_ctx(config->sched_ctxs[s].id);
+				nready += starpu_sched_ctx_get_nready_tasks(config->sched_ctxs[s].id);
 			}
 		}
 	}
@@ -855,11 +1048,6 @@ int starpu_task_nready(void)
 	return nready;
 }
 
-void _starpu_initialize_current_task_key(void)
-{
-	STARPU_PTHREAD_KEY_CREATE(&current_task_key, NULL);
-}
-
 /* Return the task currently executed by the worker, or NULL if this is called
  * either from a thread that is not a task or simply because there is no task
  * being executed at the moment. */
@@ -873,6 +1061,31 @@ void _starpu_set_current_task(struct starpu_task *task)
 	STARPU_PTHREAD_SETSPECIFIC(current_task_key, task);
 }
 
+#ifdef STARPU_OPENMP
+/* Prepare the fields of the currentl task for accepting a new set of
+ * dependencies in anticipation of becoming a continuation.
+ *
+ * When the task becomes 'continued', it will only be queued again when the new
+ * set of dependencies is fulfilled. */
+void _starpu_task_prepare_for_continuation(void)
+{
+	_starpu_job_prepare_for_continuation(_starpu_get_job_associated_to_task(starpu_task_get_current()));
+}
+
+void _starpu_task_prepare_for_continuation_ext(unsigned continuation_resubmit,
+		void (*continuation_callback_on_sleep)(void *arg), void *continuation_callback_on_sleep_arg)
+{
+	_starpu_job_prepare_for_continuation_ext(_starpu_get_job_associated_to_task(starpu_task_get_current()),
+		continuation_resubmit, continuation_callback_on_sleep, continuation_callback_on_sleep_arg);
+}
+
+void _starpu_task_set_omp_cleanup_callback(struct starpu_task *task, void (*omp_cleanup_callback)(void *arg), void *omp_cleanup_callback_arg)
+{
+	_starpu_job_set_omp_cleanup_callback(_starpu_get_job_associated_to_task(task),
+		omp_cleanup_callback, omp_cleanup_callback_arg);
+}
+#endif
+
 /*
  * Returns 0 if tasks does not use any multiformat handle, 1 otherwise.
  */
@@ -880,7 +1093,8 @@ int
 _starpu_task_uses_multiformat_handles(struct starpu_task *task)
 {
 	unsigned i;
-	for (i = 0; i < task->cl->nbuffers; i++)
+	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+	for (i = 0; i < nbuffers; i++)
 	{
 		if (_starpu_data_is_multiformat_handle(STARPU_TASK_GET_HANDLE(task, i)))
 			return 1;
@@ -918,6 +1132,8 @@ _starpu_handle_needs_conversion_task_for_arch(starpu_data_handle_t handle,
 					return 0;
 				case STARPU_CUDA_RAM:      /* Fall through */
 				case STARPU_OPENCL_RAM:
+				case STARPU_MIC_RAM:
+				case STARPU_SCC_RAM:
 					return 1;
 				default:
 					STARPU_ABORT();
@@ -925,12 +1141,16 @@ _starpu_handle_needs_conversion_task_for_arch(starpu_data_handle_t handle,
 			break;
 		case STARPU_CUDA_RAM:    /* Fall through */
 		case STARPU_OPENCL_RAM:
+		case STARPU_MIC_RAM:
+		case STARPU_SCC_RAM:
 			switch(starpu_node_get_kind(handle->mf_node))
 			{
 				case STARPU_CPU_RAM:
 					return 1;
 				case STARPU_CUDA_RAM:
 				case STARPU_OPENCL_RAM:
+				case STARPU_MIC_RAM:
+				case STARPU_SCC_RAM:
 					return 0;
 				default:
 					STARPU_ABORT();
@@ -943,21 +1163,6 @@ _starpu_handle_needs_conversion_task_for_arch(starpu_data_handle_t handle,
 	return -EINVAL;
 }
 
-starpu_cpu_func_t _starpu_task_get_cpu_nth_implementation(struct starpu_codelet *cl, unsigned nimpl)
-{
-	return cl->cpu_funcs[nimpl];
-}
-
-starpu_cuda_func_t _starpu_task_get_cuda_nth_implementation(struct starpu_codelet *cl, unsigned nimpl)
-{
-	return cl->cuda_funcs[nimpl];
-}
-
-starpu_opencl_func_t _starpu_task_get_opencl_nth_implementation(struct starpu_codelet *cl, unsigned nimpl)
-{
-	return cl->opencl_funcs[nimpl];
-}
-
 void starpu_task_set_implementation(struct starpu_task *task, unsigned impl)
 {
 	_starpu_get_job_associated_to_task(task)->nimpl = impl;
@@ -968,36 +1173,21 @@ unsigned starpu_task_get_implementation(struct starpu_task *task)
 	return _starpu_get_job_associated_to_task(task)->nimpl;
 }
 
-void _starpu_sleep(struct timespec ts)
-{
-#ifdef STARPU_HAVE_WINDOWS
-	Sleep((ts.tv_sec * 1000) + (ts.tv_nsec / 1000000));
-#else
-	struct timespec req, rem;
-
-	req = ts;
-	while (nanosleep(&req, &rem))
-		req = rem;
-#endif
-}
-
 static starpu_pthread_t watchdog_thread;
 
 /* Check from times to times that StarPU does finish some tasks */
-static void *watchdog_func(void *foo STARPU_ATTRIBUTE_UNUSED)
+static void *watchdog_func(void *arg)
 {
-	struct timespec ts;
-	char *timeout_env;
-	unsigned long long timeout;
+	char *timeout_env = arg;
+	float timeout;
 
-	if (! (timeout_env = getenv("STARPU_WATCHDOG_TIMEOUT")))
-		return NULL;
-
-	timeout = atoll(timeout_env);
-	ts.tv_sec = timeout / 1000000;
-	ts.tv_nsec = (timeout % 1000000) * 1000;
+#ifdef _MSC_VER
+	timeout = ((float) _atoi64(timeout_env)) / 1000000;
+#else
+	timeout = ((float) atoll(timeout_env)) / 1000000;
+#endif
 	struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
-	
+
 	STARPU_PTHREAD_MUTEX_LOCK(&config->submitted_mutex);
 	while (_starpu_machine_is_running())
 	{
@@ -1005,21 +1195,38 @@ static void *watchdog_func(void *foo STARPU_ATTRIBUTE_UNUSED)
 		config->watchdog_ok = 0;
 		STARPU_PTHREAD_MUTEX_UNLOCK(&config->submitted_mutex);
 
-		_starpu_sleep(ts);
+		/* If we do a sleep(timeout), we might have to wait too long at the end of the computation. */
+		/* To avoid that, we do several sleep() of 1s (and check after each if starpu is still running) */
+		float t;
+		for (t = timeout ; t > 1.; t--)
+		{
+			starpu_sleep(1.);
+			if (!_starpu_machine_is_running())
+			{
+				/* Application finished, don't bother finishing the sleep */
+				STARPU_PTHREAD_MUTEX_UNLOCK(&config->submitted_mutex);
+				return NULL;
+			}
+		}
+		/* and one final sleep (of less than 1 s) with the rest (if needed) */
+		if (t > 0.)
+			starpu_sleep(t);
 
 		STARPU_PTHREAD_MUTEX_LOCK(&config->submitted_mutex);
 		if (!config->watchdog_ok && last_nsubmitted
 				&& last_nsubmitted == starpu_task_nsubmitted())
 		{
-			fprintf(stderr,"The StarPU watchdog detected that no task finished for %u.%06us (can be configure through STARPU_WATCHDOG_TIMEOUT)\n", (unsigned)ts.tv_sec, (unsigned)ts.tv_nsec/1000);
-			if (getenv("STARPU_WATCHDOG_CRASH"))
+			fprintf(stderr,"The StarPU watchdog detected that no task finished for %fs (can be configure through STARPU_WATCHDOG_TIMEOUT)\n", timeout);
+			if (watchdog_crash)
 			{
 				fprintf(stderr,"Crashing the process\n");
-				assert(0);
+				raise(SIGABRT);
 			}
 			else
 				fprintf(stderr,"Set the STARPU_WATCHDOG_CRASH environment variable if you want to abort the process in such a case\n");
 		}
+		/* Only shout again after another period */
+		config->watchdog_ok = 1;
 	}
 	STARPU_PTHREAD_MUTEX_UNLOCK(&config->submitted_mutex);
 	return NULL;
@@ -1028,11 +1235,22 @@ static void *watchdog_func(void *foo STARPU_ATTRIBUTE_UNUSED)
 void _starpu_watchdog_init(void)
 {
 	struct _starpu_machine_config *config = (struct _starpu_machine_config *)_starpu_get_machine_config();
+	char *timeout_env = starpu_getenv("STARPU_WATCHDOG_TIMEOUT");
+
 	STARPU_PTHREAD_MUTEX_INIT(&config->submitted_mutex, NULL);
-	STARPU_PTHREAD_CREATE(&watchdog_thread, NULL, watchdog_func, NULL);
+
+	if (!timeout_env)
+		return;
+
+	STARPU_PTHREAD_CREATE(&watchdog_thread, NULL, watchdog_func, timeout_env);
 }
 
 void _starpu_watchdog_shutdown(void)
 {
+	char *timeout_env = starpu_getenv("STARPU_WATCHDOG_TIMEOUT");
+
+	if (!timeout_env)
+		return;
+
 	starpu_pthread_join(watchdog_thread, NULL);
 }
