@@ -1,7 +1,7 @@
-/* StarPU --- Runtime system for heterogeneous multicore architectures.
- *
- * Copyright (C) 2009-2014  Université de Bordeaux 1
- * Copyright (C) 2010, 2011, 2012, 2013, 2014  Centre National de la Recherche Scientifique
+/* StarPU --- Runtime system for heterogeneous multicore architectures. *
+ * Copyright (C) 2009-2016  Université de Bordeaux
+ * Copyright (C) 2010, 2011, 2012, 2013, 2014, 2015, 2016  CNRS
+ * Copyright (C) 2014  INRIA
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -19,10 +19,18 @@
 #include <datawizard/coherency.h>
 #include <datawizard/copy_driver.h>
 #include <datawizard/write_back.h>
+#include <datawizard/memory_nodes.h>
 #include <core/dependencies/data_concurrency.h>
+#include <core/disk.h>
 #include <profiling/profiling.h>
 #include <math.h>
 #include <core/task.h>
+#include <starpu_scheduler.h>
+#include <core/workers.h>
+
+#ifdef STARPU_SIMGRID
+#include <core/simgrid.h>
+#endif
 
 static int link_supports_direct_transfers(starpu_data_handle_t handle, unsigned src_node, unsigned dst_node, unsigned *handling_node);
 int _starpu_select_src_node(starpu_data_handle_t handle, unsigned destination)
@@ -35,9 +43,9 @@ int _starpu_select_src_node(starpu_data_handle_t handle, unsigned destination)
 	/* first find a valid copy, either a STARPU_OWNER or a STARPU_SHARED */
 	unsigned node;
 
-	unsigned src_node_mask = 0;
 	size_t size = _starpu_data_get_size(handle);
 	double cost = INFINITY;
+	unsigned src_node_mask = 0;
 
 	for (node = 0; node < nnodes; node++)
 	{
@@ -55,7 +63,7 @@ int _starpu_select_src_node(starpu_data_handle_t handle, unsigned destination)
 	}
 
 	/* we should have found at least one copy ! */
-	STARPU_ASSERT_MSG(src_node_mask != 0, "The data for this handle is requested, but this handle does not have a valid value. Perhaps some initialization task is missing?");
+	STARPU_ASSERT_MSG(src_node_mask != 0, "The data for the handle %p is requested, but the handle does not have a valid value. Perhaps some initialization task is missing?", handle);
 
 	/* Without knowing the size, we won't know the cost */
 	if (!size)
@@ -89,32 +97,73 @@ int _starpu_select_src_node(starpu_data_handle_t handle, unsigned destination)
 		}
 
 	if (cost && src_node != -1)
+	{
 		/* Could estimate through cost, return that */
+		STARPU_ASSERT(handle->per_node[src_node].allocated);
+		STARPU_ASSERT(handle->per_node[src_node].initialized);
 		return src_node;
+	}
+	
+	int i_ram = -1;
+	int i_gpu = -1;
+	int i_disk = -1;
 
 	/* Revert to dumb strategy: take RAM unless only a GPU has it */
 	for (i = 0; i < nnodes; i++)
 	{
+		
 		if (src_node_mask & (1<<i))
 		{
-			/* this is a potential candidate */
-			src_node = i;
+			int (*can_copy)(void *src_interface, unsigned src_node, void *dst_interface, unsigned dst_node, unsigned handling_node) = handle->ops->copy_methods->can_copy;
+			/* Avoid transfers which the interface does not want */
+			if (can_copy)
+			{
+				void *src_interface = handle->per_node[i].data_interface;
+				void *dst_interface = handle->per_node[destination].data_interface;
+				unsigned handling_node;
+
+				if (!link_supports_direct_transfers(handle, i, destination, &handling_node))
+				{
+					/* Avoid through RAM if the interface does not want it */
+					void *ram_interface = handle->per_node[STARPU_MAIN_RAM].data_interface;
+					if ((!can_copy(src_interface, i, ram_interface, STARPU_MAIN_RAM, i)
+					  && !can_copy(src_interface, i, ram_interface, STARPU_MAIN_RAM, STARPU_MAIN_RAM))
+					 || (!can_copy(ram_interface, STARPU_MAIN_RAM, dst_interface, destination, STARPU_MAIN_RAM)
+					  && !can_copy(ram_interface, STARPU_MAIN_RAM, dst_interface, destination, destination)))
+						continue;
+				}
+			}
 
 			/* however GPU are expensive sources, really !
-			 * 	Unless peer transfer is supported.
+			 * 	Unless peer transfer is supported (and it would then have been selected above).
 			 * 	Other should be ok */
 
-			if (
-#ifndef HAVE_CUDA_MEMCPY_PEER
-					starpu_node_get_kind(i) != STARPU_CUDA_RAM &&
-#endif
-					starpu_node_get_kind(i) != STARPU_OPENCL_RAM)
-				break ;
+			if (starpu_node_get_kind(i) == STARPU_CUDA_RAM ||
+			    starpu_node_get_kind(i) == STARPU_OPENCL_RAM ||
+			    starpu_node_get_kind(i) == STARPU_MIC_RAM)
+				i_gpu = i;
+
+			if (starpu_node_get_kind(i) == STARPU_CPU_RAM || 
+			    starpu_node_get_kind(i) == STARPU_SCC_RAM ||
+			    starpu_node_get_kind(i) == STARPU_SCC_SHM)
+				i_ram = i;
+			if (starpu_node_get_kind(i) == STARPU_DISK_RAM)			
+				i_disk = i;
 		}
 	}
+	
+	/* we have to use cpu_ram in first */
+	if (i_ram != -1)
+		src_node = i_ram;
+	/* no luck we have to use the disk memory */
+	else if (i_gpu != -1)
+		src_node = i_gpu;
+	else
+		src_node = i_disk;
 
 	STARPU_ASSERT(src_node != -1);
-
+	STARPU_ASSERT(handle->per_node[src_node].allocated);
+	STARPU_ASSERT(handle->per_node[src_node].initialized);
 	return src_node;
 }
 
@@ -139,9 +188,15 @@ void _starpu_update_data_state(starpu_data_handle_t handle,
 		/* the requesting node now has the only valid copy */
 		unsigned node;
 		for (node = 0; node < nnodes; node++)
+		{
+			_STARPU_TRACE_DATA_INVALIDATE(handle, node);
 			handle->per_node[node].state = STARPU_INVALID;
+		}
 
 		requesting_replicate->state = STARPU_OWNER;
+		if (handle->home_node != -1 && handle->per_node[handle->home_node].state == STARPU_INVALID)
+			/* Notify that this MC is now dirty */
+			_starpu_memchunk_dirty(requesting_replicate->mc, requesting_replicate->memory_node);
 	}
 	else
 	{ /* read only */
@@ -162,6 +217,10 @@ void _starpu_update_data_state(starpu_data_handle_t handle,
 
 static int worker_supports_direct_access(unsigned node, unsigned handling_node)
 {
+	/* only support disk <-> ram and disk <-> disk */
+	if (starpu_node_get_kind(node) == STARPU_DISK_RAM || starpu_node_get_kind(handling_node) == STARPU_DISK_RAM)
+		return 0;
+
 	if (node == handling_node)
 		return 1;
 
@@ -173,19 +232,38 @@ static int worker_supports_direct_access(unsigned node, unsigned handling_node)
 	switch (type)
 	{
 		case STARPU_CUDA_RAM:
-#ifdef HAVE_CUDA_MEMCPY_PEER
 		{
-			enum starpu_node_kind kind = starpu_node_get_kind(handling_node);
 			/* GPUs not always allow direct remote access: if CUDA4
 			 * is enabled, we allow two CUDA devices to communicate. */
+#ifdef STARPU_SIMGRID
+			if (starpu_node_get_kind(handling_node) == STARPU_CUDA_RAM)
+			{
+				char name[16];
+				msg_host_t host;
+				const char* cuda_memcpy_peer;
+				snprintf(name, sizeof(name), "CUDA%d", _starpu_memory_node_get_devid(handling_node));
+				host = _starpu_simgrid_get_host_by_name(name);
+				cuda_memcpy_peer = MSG_host_get_property_value(host, "memcpy_peer");
+				return cuda_memcpy_peer && atoll(cuda_memcpy_peer);
+			}
+			else
+				return 0;
+#elif defined(HAVE_CUDA_MEMCPY_PEER)
+			/* simgrid */
+			enum starpu_node_kind kind = starpu_node_get_kind(handling_node);
 			return kind == STARPU_CUDA_RAM;
-		}
-#else
+#else /* HAVE_CUDA_MEMCPY_PEER */
 			/* Direct GPU-GPU transfers are not allowed in general */
 			return 0;
-#endif
+#endif /* HAVE_CUDA_MEMCPY_PEER */
+		}
 		case STARPU_OPENCL_RAM:
 			return 0;
+		case STARPU_MIC_RAM:
+			/* We don't handle direct MIC-MIC transfers yet */
+			return 0;
+		case STARPU_SCC_RAM:
+			return 1;
 		default:
 			return 1;
 	}
@@ -193,33 +271,48 @@ static int worker_supports_direct_access(unsigned node, unsigned handling_node)
 
 static int link_supports_direct_transfers(starpu_data_handle_t handle, unsigned src_node, unsigned dst_node, unsigned *handling_node)
 {
-	(void) handle; // unused
+	int (*can_copy)(void *src_interface, unsigned src_node, void *dst_interface, unsigned dst_node, unsigned handling_node) = handle->ops->copy_methods->can_copy;
+	void *src_interface = handle->per_node[src_node].data_interface;
+	void *dst_interface = handle->per_node[dst_node].data_interface;
 
-	/* XXX That's a hack until we get cudaMemcpy3DPeerAsync to work !
+	/* XXX That's a hack until we fix cudaMemcpy3DPeerAsync in the block interface
 	 * Perhaps not all data interface provide a direct GPU-GPU transfer
 	 * method ! */
 #if defined(STARPU_USE_CUDA) || defined(STARPU_SIMGRID)
 	if (src_node != dst_node && starpu_node_get_kind(src_node) == STARPU_CUDA_RAM && starpu_node_get_kind(dst_node) == STARPU_CUDA_RAM)
 	{
 		const struct starpu_data_copy_methods *copy_methods = handle->ops->copy_methods;
-		if (!copy_methods->cuda_to_cuda_async)
+		if (!copy_methods->cuda_to_cuda_async && !copy_methods->any_to_any)
 			return 0;
 	}
 #endif
 
 	/* Note: with CUDA, performance seems a bit better when issuing the transfer from the destination (tested without GPUDirect, but GPUDirect probably behave the same) */
-	if (worker_supports_direct_access(src_node, dst_node))
+	if (worker_supports_direct_access(src_node, dst_node) && (!can_copy || can_copy(src_interface, src_node, dst_interface, dst_node, dst_node)))
 	{
 		*handling_node = dst_node;
 		return 1;
 	}
 
-	if (worker_supports_direct_access(dst_node, src_node))
+	if (worker_supports_direct_access(dst_node, src_node) && (!can_copy || can_copy(src_interface, src_node, dst_interface, dst_node, src_node)))
 	{
 		*handling_node = src_node;
 		return 1;
 	}
 
+	/* Link between disk and ram */
+	if ((starpu_node_get_kind(src_node) == STARPU_DISK_RAM && starpu_node_get_kind(dst_node) == STARPU_CPU_RAM) ||
+	    (starpu_node_get_kind(src_node) == STARPU_CPU_RAM && starpu_node_get_kind(dst_node) == STARPU_DISK_RAM))
+	{
+		/* FIXME: not necessarily a worker :/ */
+		*handling_node = STARPU_MAIN_RAM;
+		return 1;
+	}
+
+	/* link between disk and disk, and they have the same kind */
+	if (_starpu_is_same_kind_disk(src_node, dst_node))
+		return 1;
+
 	return 0;
 }
 
@@ -231,13 +324,17 @@ static int determine_request_path(starpu_data_handle_t handle,
 				  unsigned src_node, unsigned dst_node,
 				  enum starpu_data_access_mode mode, int max_len,
 				  unsigned *src_nodes, unsigned *dst_nodes,
-				  unsigned *handling_nodes)
+				  unsigned *handling_nodes, unsigned write_invalidation)
 {
-	if (!(mode & STARPU_R))
+	if (src_node == dst_node || !(mode & STARPU_R))
 	{
+		if (write_invalidation)
+			/* The invalidation request will be enough */
+			return 0;
+
 		/* The destination node should only allocate the data, no transfer is required */
 		STARPU_ASSERT(max_len >= 1);
-		src_nodes[0] = 0; // ignored
+		src_nodes[0] = STARPU_MAIN_RAM; // ignored
 		dst_nodes[0] = dst_node;
 		handling_nodes[0] = dst_node;
 		return 1;
@@ -248,21 +345,47 @@ static int determine_request_path(starpu_data_handle_t handle,
 
 	if (!link_is_valid)
 	{
+		int (*can_copy)(void *src_interface, unsigned src_node, void *dst_interface, unsigned dst_node, unsigned handling_node) = handle->ops->copy_methods->can_copy;
+		void *src_interface = handle->per_node[src_node].data_interface;
+		void *dst_interface = handle->per_node[dst_node].data_interface;
+
 		/* We need an intermediate hop to implement data staging
 		 * through main memory. */
 		STARPU_ASSERT(max_len >= 2);
 
-		/* XXX we hardcode 0 as the RAM node ... */
-
 		/* GPU -> RAM */
 		src_nodes[0] = src_node;
-		dst_nodes[0] = 0;
-		handling_nodes[0] = src_node;
+		dst_nodes[0] = STARPU_MAIN_RAM;
+
+		if (starpu_node_get_kind(src_node) == STARPU_DISK_RAM)
+			/* Disks don't have their own driver thread */
+			handling_nodes[0] = dst_node;
+		else if (!can_copy || can_copy(src_interface, src_node, dst_interface, dst_node, src_node))
+		{
+			handling_nodes[0] = src_node;
+		}
+		else
+		{
+			STARPU_ASSERT_MSG(can_copy(src_interface, src_node, dst_interface, dst_node, dst_node), "interface %d refuses all kinds of transfers from node %u to node %u\n", handle->ops->interfaceid, src_node, dst_node);
+			handling_nodes[0] = dst_node;
+		}
 
 		/* RAM -> GPU */
-		src_nodes[1] = 0;
+		src_nodes[1] = STARPU_MAIN_RAM;
 		dst_nodes[1] = dst_node;
-		handling_nodes[1] = dst_node;
+
+		if (starpu_node_get_kind(dst_node) == STARPU_DISK_RAM)
+			/* Disks don't have their own driver thread */
+			handling_nodes[1] = src_node;
+		else if (!can_copy || can_copy(src_interface, src_node, dst_interface, dst_node, dst_node))
+		{
+			handling_nodes[1] = dst_node;
+		}
+		else
+		{
+			STARPU_ASSERT_MSG(can_copy(src_interface, src_node, dst_interface, dst_node, src_node), "interface %d refuses all kinds of transfers from node %u to node %u\n", handle->ops->interfaceid, src_node, dst_node);
+			handling_nodes[1] = src_node;
+		}
 
 		return 2;
 	}
@@ -274,7 +397,7 @@ static int determine_request_path(starpu_data_handle_t handle,
 		dst_nodes[0] = dst_node;
 		handling_nodes[0] = handling_node;
 
-#ifndef HAVE_CUDA_MEMCPY_PEER
+#if !defined(HAVE_CUDA_MEMCPY_PEER) && !defined(STARPU_SIMGRID)
 		STARPU_ASSERT(!(mode & STARPU_R) || starpu_node_get_kind(src_node) != STARPU_CUDA_RAM || starpu_node_get_kind(dst_node) != STARPU_CUDA_RAM);
 #endif
 
@@ -299,7 +422,7 @@ static struct _starpu_data_request *_starpu_search_existing_data_request(struct
 
                 /* perhaps we need to "upgrade" the request */
 		if (is_prefetch < r->prefetch)
-			_starpu_update_prefetch_status(r);
+			_starpu_update_prefetch_status(r, is_prefetch);
 
 		if (mode & STARPU_R)
 		{
@@ -352,17 +475,42 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 								  unsigned async,
 								  void (*callback_func)(void *), void *callback_arg)
 {
+	/* We don't care about commuting for data requests, that was handled before. */
+	mode &= ~STARPU_COMMUTE;
+
 	/* This function is called with handle's header lock taken */
 	_starpu_spin_checklocked(&handle->header_lock);
 
 	unsigned requesting_node = dst_replicate->memory_node;
+	unsigned nwait = 0;
 
-	if (dst_replicate->state != STARPU_INVALID)
+	if (mode & STARPU_W)
+	{
+		/* We will write to the buffer. We will have to wait for all
+		 * existing requests before the last request which will
+		 * invalidate all their results (which were possibly spurious,
+		 * e.g. too aggressive eviction).
+		 */
+		unsigned i, j;
+		unsigned nnodes = starpu_memory_nodes_get_count();
+		for (i = 0; i < nnodes; i++)
+			for (j = 0; j < nnodes; j++)
+				if (handle->per_node[i].request[j])
+					nwait++;
+		/* If the request is not detached (i.e. the caller really wants
+		 * proper ownership), no new requests will appear because a
+		 * reference will be kept on the dst replicate, which will
+		 * notably prevent data reclaiming.
+		 */
+	}
+
+	if (dst_replicate->state != STARPU_INVALID && (!nwait || is_prefetch))
 	{
 #ifdef STARPU_MEMORY_STATS
 		enum _starpu_cache_state old_state = dst_replicate->state;
 #endif
-		/* the data is already available so we can stop */
+		/* the data is already available and we don't have to wait for
+		 * any request, so we can stop */
 		_starpu_update_data_state(handle, dst_replicate, mode);
 		_starpu_msi_cache_hit(requesting_node);
 
@@ -375,7 +523,8 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 			_starpu_memory_handle_stats_shared_to_owner(handle, requesting_node);
 #endif
 
-		_starpu_memchunk_recently_used(dst_replicate->mc, requesting_node);
+		if (dst_replicate->mc)
+			_starpu_memchunk_recently_used(dst_replicate->mc, requesting_node);
 
 		_starpu_spin_unlock(&handle->header_lock);
 
@@ -389,15 +538,17 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 	_starpu_msi_cache_miss(requesting_node);
 
 	/* the only remaining situation is that the local copy was invalid */
-	STARPU_ASSERT(dst_replicate->state == STARPU_INVALID);
+	STARPU_ASSERT(dst_replicate->state == STARPU_INVALID || nwait);
 
 	/* find someone who already has the data */
-	int src_node = 0;
+	int src_node = -1;
 
 	if (mode & STARPU_R)
 	{
-		src_node = _starpu_select_src_node(handle, requesting_node);
-		STARPU_ASSERT(src_node != (int) requesting_node);
+		if (dst_replicate->state == STARPU_INVALID)
+			src_node = _starpu_select_src_node(handle, requesting_node);
+		else
+			src_node = requesting_node;
 		if (src_node < 0)
 		{
 			/* We will create it, no need to read an existing value */
@@ -409,7 +560,8 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 		/* if the data is in write only mode (and not SCRATCH or REDUX), there is no need for a source, data will be initialized by the task itself */
 		if (mode & STARPU_W)
 			dst_replicate->initialized = 1;
-		if (requesting_node == 0) {
+		if (requesting_node == STARPU_MAIN_RAM && !nwait)
+		{
 			/* And this is the main RAM, really no need for a
 			 * request, just allocate */
 			if (_starpu_allocate_memory_on_node(handle, dst_replicate, is_prefetch) == 0)
@@ -426,17 +578,20 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 		}
 	}
 
+#define MAX_REQUESTS 4
 	/* We can safely assume that there won't be more than 2 hops in the
 	 * current implementation */
-	unsigned src_nodes[4], dst_nodes[4], handling_nodes[4];
-	int nhops = determine_request_path(handle, src_node, requesting_node, mode, 4,
-					src_nodes, dst_nodes, handling_nodes);
+	unsigned src_nodes[MAX_REQUESTS], dst_nodes[MAX_REQUESTS], handling_nodes[MAX_REQUESTS];
+	/* keep one slot for the last W request, if any */
+	int write_invalidation = (mode & STARPU_W) && nwait && !is_prefetch;
+	int nhops = determine_request_path(handle, src_node, requesting_node, mode, MAX_REQUESTS,
+					   src_nodes, dst_nodes, handling_nodes, write_invalidation);
 
-	STARPU_ASSERT(nhops >= 1 && nhops <= 4);
-	struct _starpu_data_request *requests[nhops];
+	STARPU_ASSERT(nhops >= 0 && nhops <= MAX_REQUESTS-1);
+	struct _starpu_data_request *requests[nhops + write_invalidation];
 
 	/* Did we reuse a request for that hop ? */
-	int reused_requests[nhops];
+	int reused_requests[nhops + write_invalidation];
 
 	/* Construct an array with a list of requests, possibly reusing existing requests */
 	int hop;
@@ -469,7 +624,8 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 			/* Create a new request if there was no request to reuse */
 			r = _starpu_create_data_request(handle, hop_src_replicate,
 							hop_dst_replicate, hop_handling_node,
-							mode, ndeps, is_prefetch);
+							mode, ndeps, is_prefetch, 0);
+			nwait++;
 		}
 
 		requests[hop] = r;
@@ -489,7 +645,7 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 				STARPU_ASSERT(r->next_req_count <= STARPU_MAXNODES);
 			}
 		}
-		else
+		else if (!write_invalidation)
 			/* The last request will perform the callback after termination */
 			_starpu_data_request_append_callback(r, callback_func, callback_arg);
 
@@ -498,6 +654,51 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 			_starpu_spin_unlock(&r->lock);
 	}
 
+	if (write_invalidation)
+	{
+		/* Some requests were still pending, we have to add yet another
+		 * request, depending on them, which will invalidate their
+		 * result.
+		 */
+		struct _starpu_data_request *r = _starpu_create_data_request(handle, dst_replicate,
+							dst_replicate, requesting_node,
+							STARPU_W, nwait, is_prefetch, 1);
+
+		/* and perform the callback after termination */
+		_starpu_data_request_append_callback(r, callback_func, callback_arg);
+
+		/* We will write to the buffer. We will have to wait for all
+		 * existing requests before the last request which will
+		 * invalidate all their results (which were possibly spurious,
+		 * e.g. too aggressive eviction).
+		 */
+		unsigned i, j;
+		unsigned nnodes = starpu_memory_nodes_get_count();
+		for (i = 0; i < nnodes; i++)
+			for (j = 0; j < nnodes; j++)
+			{
+				struct _starpu_data_request *r2 = handle->per_node[i].request[j];
+				if (r2)
+				{
+					_starpu_spin_lock(&r2->lock);
+					if (is_prefetch < r2->prefetch)
+						/* Hasten the request we will have to wait for */
+						_starpu_update_prefetch_status(r2, is_prefetch);
+					r2->next_req[r2->next_req_count++] = r;
+					STARPU_ASSERT(r2->next_req_count <= STARPU_MAXNODES + 1);
+					_starpu_spin_unlock(&r2->lock);
+					nwait--;
+				}
+			}
+		STARPU_ASSERT(nwait == 0);
+
+		nhops++;
+		requests[nhops - 1] = r;
+		/* existing requests will post this one */
+		reused_requests[nhops - 1] = 1;
+	}
+	STARPU_ASSERT(nhops);
+
 	if (!async)
 		requests[nhops - 1]->refcnt++;
 
@@ -511,7 +712,7 @@ struct _starpu_data_request *_starpu_create_request_to_fetch_data(starpu_data_ha
 }
 
 int _starpu_fetch_data_on_node(starpu_data_handle_t handle, struct _starpu_data_replicate *dst_replicate,
-			       enum starpu_data_access_mode mode, unsigned detached, unsigned async,
+			       enum starpu_data_access_mode mode, unsigned detached, unsigned is_prefetch, unsigned async,
 			       void (*callback_func)(void *), void *callback_arg)
 {
 	unsigned local_node = _starpu_memory_node_get_local_key();
@@ -535,7 +736,7 @@ int _starpu_fetch_data_on_node(starpu_data_handle_t handle, struct _starpu_data_
 
 	struct _starpu_data_request *r;
 	r = _starpu_create_request_to_fetch_data(handle, dst_replicate, mode,
-						 detached, async, callback_func, callback_arg);
+						 is_prefetch, async, callback_func, callback_arg);
 
 	/* If no request was created, the handle was already up-to-date on the
 	 * node. In this case, _starpu_create_request_to_fetch_data has already
@@ -550,14 +751,19 @@ int _starpu_fetch_data_on_node(starpu_data_handle_t handle, struct _starpu_data_
         return ret;
 }
 
+static int idle_prefetch_data_on_node(starpu_data_handle_t handle, struct _starpu_data_replicate *replicate, enum starpu_data_access_mode mode)
+{
+	return _starpu_fetch_data_on_node(handle, replicate, mode, 1, 2, 1, NULL, NULL);
+}
+
 static int prefetch_data_on_node(starpu_data_handle_t handle, struct _starpu_data_replicate *replicate, enum starpu_data_access_mode mode)
 {
-	return _starpu_fetch_data_on_node(handle, replicate, mode, 1, 1, NULL, NULL);
+	return _starpu_fetch_data_on_node(handle, replicate, mode, 1, 1, 1, NULL, NULL);
 }
 
 static int fetch_data(starpu_data_handle_t handle, struct _starpu_data_replicate *replicate, enum starpu_data_access_mode mode)
 {
-	return _starpu_fetch_data_on_node(handle, replicate, mode, 0, 0, NULL, NULL);
+	return _starpu_fetch_data_on_node(handle, replicate, mode, 0, 0, 0, NULL, NULL);
 }
 
 uint32_t _starpu_get_data_refcnt(starpu_data_handle_t handle, unsigned node)
@@ -637,13 +843,14 @@ static void _starpu_set_data_requested_flag_if_needed(starpu_data_handle_t handl
 
 int starpu_prefetch_task_input_on_node(struct starpu_task *task, unsigned node)
 {
-	unsigned nbuffers = task->cl->nbuffers;
+	STARPU_ASSERT(!task->prefetched);
+	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
 	unsigned index;
 
 	for (index = 0; index < nbuffers; index++)
 	{
 		starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, index);
-		enum starpu_data_access_mode mode = STARPU_CODELET_GET_MODE(task->cl, index);
+		enum starpu_data_access_mode mode = STARPU_TASK_GET_MODE(task, index);
 
 		if (mode & (STARPU_SCRATCH|STARPU_REDUX))
 			continue;
@@ -657,16 +864,48 @@ int starpu_prefetch_task_input_on_node(struct starpu_task *task, unsigned node)
 	return 0;
 }
 
+int starpu_idle_prefetch_task_input_on_node(struct starpu_task *task, unsigned node)
+{
+	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+	unsigned index;
+
+	for (index = 0; index < nbuffers; index++)
+	{
+		starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, index);
+		enum starpu_data_access_mode mode = STARPU_TASK_GET_MODE(task, index);
+
+		if (mode & (STARPU_SCRATCH|STARPU_REDUX))
+			continue;
+
+		struct _starpu_data_replicate *replicate = &handle->per_node[node];
+		idle_prefetch_data_on_node(handle, replicate, mode);
+	}
+
+	return 0;
+}
+
+
 static struct _starpu_data_replicate *get_replicate(starpu_data_handle_t handle, enum starpu_data_access_mode mode, int workerid, unsigned node)
 {
 	if (mode & (STARPU_SCRATCH|STARPU_REDUX))
+	{
+		STARPU_ASSERT(workerid >= 0);
+		if (!handle->per_worker)
+		{
+			_starpu_spin_lock(&handle->header_lock);
+			if (!handle->per_worker)
+				_starpu_data_initialize_per_worker(handle);
+			_starpu_spin_unlock(&handle->header_lock);
+		}
 		return &handle->per_worker[workerid];
+	}
 	else
 		/* That's a "normal" buffer (R/W) */
 		return &handle->per_node[node];
 }
 
-int _starpu_fetch_task_input(struct _starpu_job *j, uint32_t mask)
+/* Synchronously fetch data for a given task (if it's not there already) */
+int _starpu_fetch_task_input(struct _starpu_job *j)
 {
 	_STARPU_TRACE_START_FETCH_INPUT(NULL);
 
@@ -676,12 +915,16 @@ int _starpu_fetch_task_input(struct _starpu_job *j, uint32_t mask)
 		_starpu_clock_gettime(&task->profiling_info->acquire_data_start_time);
 
 	struct _starpu_data_descr *descrs = _STARPU_JOB_GET_ORDERED_BUFFERS(j);
-	unsigned nbuffers = task->cl->nbuffers;
+	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
 
 	unsigned local_memory_node = _starpu_memory_node_get_local_key();
 
 	int workerid = starpu_worker_get_id();
 
+#ifdef STARPU_USE_FXT
+	unsigned long total_size = 0;
+#endif
+
 	unsigned index;
 	for (index = 0; index < nbuffers; index++)
 	{
@@ -691,6 +934,10 @@ int _starpu_fetch_task_input(struct _starpu_job *j, uint32_t mask)
 		int node = descrs[index].node;
 		if (node == -1)
 			node = local_memory_node;
+		if (mode == STARPU_NONE ||
+			(mode & ((1<<STARPU_MODE_SHIFT) - 1)) >= STARPU_ACCESS_MODE_MAX ||
+			(mode >> STARPU_MODE_SHIFT) >= (STARPU_SHIFTED_MODE_MAX >> STARPU_MODE_SHIFT))
+			STARPU_ASSERT_MSG(0, "mode %d (0x%x) is bogus\n", mode, mode);
 
 		struct _starpu_data_replicate *local_replicate;
 
@@ -705,13 +952,18 @@ int _starpu_fetch_task_input(struct _starpu_job *j, uint32_t mask)
 		ret = fetch_data(handle, local_replicate, mode);
 		if (STARPU_UNLIKELY(ret))
 			goto enomem;
+
+#ifdef STARPU_USE_FXT
+		total_size += _starpu_data_get_size(handle);
+#endif
 	}
 
+	_STARPU_TRACE_DATA_LOAD(workerid,total_size);
 	/* Now that we have taken the data locks in locking order, fill the codelet interfaces in function order.  */
 	for (index = 0; index < nbuffers; index++)
 	{
 		starpu_data_handle_t handle = STARPU_TASK_GET_HANDLE(task, index);
-		enum starpu_data_access_mode mode = STARPU_CODELET_GET_MODE(task->cl, index);
+		enum starpu_data_access_mode mode = STARPU_TASK_GET_MODE(task, index);
 		int node = descrs[index].node;
 		if (node == -1)
 			node = local_memory_node;
@@ -758,23 +1010,25 @@ enomem:
 
 		local_replicate = get_replicate(handle, mode, workerid, node);
 
-		_starpu_release_data_on_node(handle, mask, local_replicate);
+		_starpu_release_data_on_node(handle, 0, local_replicate);
 	}
 
 	return -1;
 }
 
-void _starpu_push_task_output(struct _starpu_job *j, uint32_t mask)
+/* Release task data dependencies */
+void __starpu_push_task_output(struct _starpu_job *j)
 {
-	_STARPU_TRACE_START_PUSH_OUTPUT(NULL);
-
+#ifdef STARPU_OPENMP
+	STARPU_ASSERT(!j->continuation);
+#endif
 	int profiling = starpu_profiling_status_get();
 	struct starpu_task *task = j->task;
 	if (profiling && task->profiling_info)
 		_starpu_clock_gettime(&task->profiling_info->release_data_start_time);
 
         struct _starpu_data_descr *descrs = _STARPU_JOB_GET_ORDERED_BUFFERS(j);
-        unsigned nbuffers = task->cl->nbuffers;
+        unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
 
 	int workerid = starpu_worker_get_id();
 	unsigned local_memory_node = _starpu_memory_node_get_local_key();
@@ -785,7 +1039,7 @@ void _starpu_push_task_output(struct _starpu_job *j, uint32_t mask)
 		starpu_data_handle_t handle = descrs[index].handle;
 		enum starpu_data_access_mode mode = descrs[index].mode;
 		int node = descrs[index].node;
-		if (node == -1)
+		if (node == -1 && task->cl->where != STARPU_NOWHERE)
 			node = local_memory_node;
 
 		struct _starpu_data_replicate *local_replicate;
@@ -796,23 +1050,125 @@ void _starpu_push_task_output(struct _starpu_job *j, uint32_t mask)
 			 * _starpu_compar_handles */
 			continue;
 
-		local_replicate = get_replicate(handle, mode, workerid, node);
+		if (node != -1)
+			local_replicate = get_replicate(handle, mode, workerid, node);
 
 		/* Keep a reference for future
 		 * _starpu_release_task_enforce_sequential_consistency call */
 		_starpu_spin_lock(&handle->header_lock);
 		handle->busy_count++;
-		_starpu_spin_unlock(&handle->header_lock);
 
-		_starpu_release_data_on_node(handle, mask, local_replicate);
+		if (node == -1)
+		{
+			/* NOWHERE case, just notify dependencies */
+			if (!_starpu_notify_data_dependencies(handle))
+				_starpu_spin_unlock(&handle->header_lock);
+		}
+		else
+		{
+			_starpu_spin_unlock(&handle->header_lock);
+			_starpu_release_data_on_node(handle, 0, local_replicate);
+		}
 	}
 
 	if (profiling && task->profiling_info)
 		_starpu_clock_gettime(&task->profiling_info->release_data_end_time);
+}
 
+/* Version for a driver running on a worker: we show the driver state in the trace */
+void _starpu_push_task_output(struct _starpu_job *j)
+{
+	_STARPU_TRACE_START_PUSH_OUTPUT(NULL);
+	__starpu_push_task_output(j);
 	_STARPU_TRACE_END_PUSH_OUTPUT(NULL);
 }
 
+struct fetch_nowhere_wrapper
+{
+	struct _starpu_job *j;
+	unsigned pending;
+};
+
+static void _starpu_fetch_nowhere_task_input_cb(void *arg);
+/* Asynchronously fetch data for a task which will have no content */
+void _starpu_fetch_nowhere_task_input(struct _starpu_job *j)
+{
+	int profiling = starpu_profiling_status_get();
+	struct starpu_task *task = j->task;
+	if (profiling && task->profiling_info)
+		_starpu_clock_gettime(&task->profiling_info->acquire_data_start_time);
+
+	struct _starpu_data_descr *descrs = _STARPU_JOB_GET_ORDERED_BUFFERS(j);
+	unsigned nbuffers = STARPU_TASK_GET_NBUFFERS(task);
+	unsigned nfetchbuffers = 0;
+	struct fetch_nowhere_wrapper *wrapper;
+
+	unsigned index;
+	for (index = 0; index < nbuffers; index++)
+	{
+		int node = descrs[index].node;
+		if (node != -1)
+			nfetchbuffers++;
+	}
+
+	if (!nfetchbuffers)
+	{
+		/* Nothing to fetch actually, already finished! */
+		__starpu_push_task_output(j);
+		_starpu_handle_job_termination(j);
+		_STARPU_LOG_OUT_TAG("handle_job_termination");
+		return;
+	}
+
+	wrapper = malloc(sizeof(*wrapper));
+	wrapper->j = j;
+	wrapper->pending = nfetchbuffers;
+
+	for (index = 0; index < nbuffers; index++)
+	{
+		starpu_data_handle_t handle = descrs[index].handle;
+		enum starpu_data_access_mode mode = descrs[index].mode;
+		int node = descrs[index].node;
+		if (node == -1)
+			continue;
+
+		if (mode == STARPU_NONE ||
+			(mode & ((1<<STARPU_MODE_SHIFT) - 1)) >= STARPU_ACCESS_MODE_MAX ||
+			(mode >> STARPU_MODE_SHIFT) >= (STARPU_SHIFTED_MODE_MAX >> STARPU_MODE_SHIFT))
+			STARPU_ASSERT_MSG(0, "mode %d (0x%x) is bogus\n", mode, mode);
+		STARPU_ASSERT(mode != STARPU_SCRATCH && mode != STARPU_REDUX);
+
+		struct _starpu_data_replicate *local_replicate;
+
+		local_replicate = get_replicate(handle, mode, -1, node);
+
+		_starpu_fetch_data_on_node(handle, local_replicate, mode, 0, 0, 1, _starpu_fetch_nowhere_task_input_cb, wrapper);
+	}
+
+	if (profiling && task->profiling_info)
+		_starpu_clock_gettime(&task->profiling_info->acquire_data_end_time);
+}
+
+static void _starpu_fetch_nowhere_task_input_cb(void *arg)
+{
+	/* One more transfer finished */
+	struct fetch_nowhere_wrapper *wrapper = arg;
+
+	unsigned pending = STARPU_ATOMIC_ADD(&wrapper->pending, -1);
+	ANNOTATE_HAPPENS_BEFORE(&wrapper->pending);
+	if (pending == 0)
+	{
+		ANNOTATE_HAPPENS_AFTER(&wrapper->pending);
+
+		/* Finished transferring, task is over */
+		struct _starpu_job *j = wrapper->j;
+		free(wrapper);
+		__starpu_push_task_output(j);
+		_starpu_handle_job_termination(j);
+		_STARPU_LOG_OUT_TAG("handle_job_termination");
+	}
+}
+
 /* NB : this value can only be an indication of the status of a data
 	at some point, but there is no strong garantee ! */
 unsigned _starpu_is_data_present_or_requested(starpu_data_handle_t handle, unsigned node)
