@@ -1,6 +1,6 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2012  Université de Bordeaux 1
+ * Copyright (C) 2012, 2015  Université de Bordeaux
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -31,18 +31,27 @@ const char *_starpu_codelet_get_model_name(struct starpu_codelet *cl)
 
 const char *_starpu_job_get_model_name(struct _starpu_job *j)
 {
-	const char *ret = NULL;
-
 	if (!j)
 		return NULL;
 
 	struct starpu_task *task = j->task;
-	if (task)
-		ret = _starpu_codelet_get_model_name(task->cl);
+	if (!task)
+		return NULL;
 
-#ifdef STARPU_USE_FXT
-	if (!ret)
-		ret = j->model_name;
-#endif
-	return ret;
+	return _starpu_codelet_get_model_name(task->cl);
+}
+
+const char *_starpu_job_get_task_name(struct _starpu_job *j)
+{
+	if (!j)
+		return NULL;
+
+	struct starpu_task *task = j->task;
+	if (!task)
+		return NULL;
+
+	if (task->name)
+		return task->name;
+	else
+		return _starpu_job_get_model_name(j);
 }
